
:- module(trie_sp, [
	      trie_to_bdd/3,
	      trie_to_bdd_tree/2,
	tabled_trie_to_bdd/3]).

:- use_module((bdd)).
:- use_module(library(tries)).
:- use_module(library(rbtrees)).

trie_to_formula(Trie,or(V,Disjs),Map0,Map) :-
    trie_get_first_entry(Trie, E),
    trie_get_last_entry(Trie, EF),
    trie_get_entry(E, V0),
    form(V0,V,Map0,MapI),
    collect(E,EF,Disjs,MapI,Map).

collect(E,E,0,Map,Map) :- !.
collect(E,EF,or(V,Disjs),Map0,Map) :-
    trie_get_entry(E, V0),
    form(V0,V,Map0,MapI),
    trie_traverse_next(E, E1),
    collect(E1,EF,Disjs,MapI,Map).

form([],1,Map,Map).
form([not(A)|As],and(not(V),Vs),Map0,Map) :-
    !,
    check(Map0,A,V,MapI),
    form(As,Vs,MapI,Map).
form([A|As],and(V,Vs),Map0,Map) :-
    check(Map0,A,V,MapI),
    form(As,Vs,MapI,Map).




trie_to_bdd_tree(Trie, Tree) :-
    rb_new(Map0),
    trie_to_formula(Trie, Formula,Map0,Map),
    rb_visit(Map, MapList),
    extract_vars(MapList, Vs),
    bdd_new(Formula, Vs, BDD),
    bdd_etree(BDD, MapList,Tree),
    bdd_close(BDD).

tabled_trie_to_bdd(Trie, BDD, MapList) :-
    trie_to_bdd(Trie, BDD, MapList).
/*
	trie_to_list(Trie, Complex),
	rb_new(Map0),
	rb_new(Tab0),
	Complex = [list(Els)],
	tabled_complex_to_andor(Els,Map0,Map,Tab0,_Tab,Tree),
	rb_visit(Map, MapList),
	extract_vars(MapList, Vs),
	bdd_new(Tree, Vs, BDD),
	bdd_tree(BDD, bdd(_, L, _)), length(L,Len), writeln(Len).
	*/


extract_vars([], []).
extract_vars([(_-V)|MapList], [V|Vs]) :-
	extract_vars(MapList, Vs).

complex_to_andor(empty, Map, Map, 0).
complex_to_andor([list(El)], Map0, MapF, T) :-
    !,
    complex_to_and(El, Map0, MapF, T).
complex_to_andor([list(Els),list(Els1)], Map0, MapF, or(T1,T2)) :-
    !,
    complex_to_and(Els, Map0, MapI, T1),
    complex_to_and(Els1, MapI, MapF, T2).
complex_to_andor([list(Els),Els1|Lists], Map0, MapF, or(T1,T2)) :-
    !,
    complex_to_and(Els, Map0, MapI, T1),
    complex_to_andor([Els1|Lists], MapI, MapF, T2).
complex_to_andor(Els, Map0, MapF, T) :-
    complex_to_and(Els, Map0, MapF, T).

complex_to_and([El1,El2|REls], Map0, MapF, and(T1,T2)) :-
    !,
    complex_to_node(El1, Map0, MapI, T1),
    complex_to_and([El2|REls], MapI, MapF, T2).
complex_to_and([El1,El2], Map0, MapF, and(T1,T2)) :-
    !,
    complex_to_node(El1, Map0, MapI, T1),
    complex_to_node(El2, MapI, MapF, T2).
complex_to_and([El], Map0, MapF, T) :-
    complex_to_node(El, Map0, MapF, T).

complex_to_node([El|Els], Map0, MapF,T2) :-  !,
	complex_to_andor([El|Els], Map0, MapF, T2).
complex_to_node(int(A1,[endlist]), Map0, MapF, V) :- !,
	check(Map0, A1, V, MapF).
complex_to_node(atom(true,[endlist]), Map, Map, 1) :- !.
complex_to_node(atom(false,[endlist]), Map, Map, 1) :- !.
complex_to_node(atom(A1,[endlist]), Map0, MapF, V) :- !,
	check(Map0, A1, V, MapF).
complex_to_node(functor(not,1,[int(A1,[endlist])]), Map0, MapF, not(V)) :- !,
    check(Map0, A1, V, MapF).
complex_to_node(functor(not,1,[atom(A1,[endlist])]), Map0, MapF, not(V)) :- !,
	check(Map0, A1, V, MapF).
complex_to_node(int(A1,Els), Map0, MapF, and(V,T2)) :-  !,
	check(Map0, A1, V, MapI),
	complex_to_andor(Els, MapI, MapF, T2).
complex_to_node(atom(A1,Els), Map0, MapF, and(V,T2)) :-  !,
	check(Map0, A1, V, MapI),
	complex_to_andor(Els, MapI, MapF, T2).
complex_to_node(functor(not,1,[int(A1,Els)]), Map0, MapF, and(not(V),T2)) :- !,
	check(Map0, A1, V, MapI),
	complex_to_andor(Els, MapI, MapF, T2).
complex_to_node(functor(not,1,[atom(A1,Els)]), Map0, MapF, and(not(V),T2)) :- !,
	check(Map0, A1, V, MapI),
	complex_to_andor(Els, MapI, MapF, T2).
% HASH TABLE, it can be an OR or an AND.
complex_to_node(functor(not,1,[int(A1,Els)|More]), Map0, MapF, or(NOTV1,O2)) :-
	check(Map0, A1, V, MapI),
	(Els == [endlist]
	->
	  NOTV1 = not(V),
	  MapI = MapI2
	;
	  complex_to_andor(Els, MapI, MapI2, T2),
	  NOTV1 = and(not(V), T2)
	),
	complex_to_and(functor(not,1,More), MapI2, MapF, O2).
complex_to_node(functor(not,1,[atom(A1,Els)|More]), Map0, MapF, or(NOTV1,O2)) :-
	check(Map0, A1, V, MapI),
	(Els == [endlist]
	->
	  NOTV1 = not(V),
	  MapI = MapI2
	;
	  complex_to_andor(Els, MapI, MapI2, T2),
	  NOTV1 = and(not(V), T2)
	),
	complex_to_and(functor(not,1,More), MapI2, MapF, O2).

tabled_complex_to_andor(empty, Map, Map, Tab, Tab, 0).
tabled_complex_to_andor(T, Map, Map, Tab, Tab, V) :-
	rb_lookup(T, V, Tab), !,
	increment_ref_count(V).
tabled_complex_to_andor(IN, Map, Map, Tab, Tab, 1) :-
	IN =  !.
tabled_complex_to_andor([Els], Map0, MapF, Tab0, TabF, V) :-
	tabled_complex_to_and(Els, Map0, MapF, Tab0, TabF, V).
tabled_complex_to_andor([El1,El2|Els], Map0, MapF, Tab0, TabF, or(T1,T2)) :-
	tabled_complex_to_and(El1, Map0, MapI, Tab0, TabI, T1),
	tabled_complex_to_andor([El2|Els], MapI, MapF, TabI, TabF, T2).

tabled_complex_to_and(int(A1,[endlist]), Map0, MapF, Tab, Tab, V) :- !,
	check(Map0, A1, V, MapF).
tabled_complex_to_and(atom(A1,[endlist]), Map0, MapF, Tab, Tab, V) :- !,
	check(Map0, A1, V, MapF).
tabled_complex_to_and(functor(not,1,[int(A1,[endlist])]), Map0, MapF, Tab, Tab, not(V)) :- !,
	check(Map0, A1, V, MapF).
tabled_complex_to_and(functor(not,1,[atom(A1,[endlist])]), Map0, MapF, Tab, Tab, not(V)) :- !,
	check(Map0, A1, V, MapF).
tabled_complex_to_and(T, Map, Map, Tab, Tab, V) :-
	rb_lookup(T, V, Tab), !,
	increment_ref_count(V).
tabled_complex_to_and(IN, Map0, MapF, Tab0, TabF, OUT) :-
	IN = int(A1,Els), !,
	check(Map0, A1, V, MapI),
	rb_insert(Tab0, IN, OUT, TabI),
	OUT = and(0, _, V, T2),
	tabled_complex_to_andor(Els, MapI, MapF, TabI, TabF, T2).
tabled_complex_to_and(IN, Map0, MapF, Tab0, TabF, OUT) :-
	IN = atom(A1,Els), !,
	check(Map0, A1, V, MapI),
	rb_insert(Tab0, IN, OUT, TabI),
	OUT = and(0, _, V, T2),
	tabled_complex_to_andor(Els, MapI, MapF, TabI, TabF, T2).
tabled_complex_to_and(IN, Map0, MapF, Tab0, TabF, OUT) :-
	IN = functor(not,1,[int(A1,Els)]), !,
	check(Map0, A1, V, MapI),
	rb_insert(Tab0, IN, OUT, TabI),
	OUT = and(0, _, not(V), T2),
	tabled_complex_to_andor(Els, MapI, MapF, TabI, TabF, T2).
tabled_complex_to_and(IN, Map0, MapF, Tab0, TabF, OUT) :-
	IN = functor(not,1,[atom(A1,Els)]), !,
	check(Map0, A1, V, MapI),
	rb_insert(Tab0, IN, OUT, TabI),
	OUT = and(0, _, not(V), T2),
	tabled_complex_to_andor(Els, MapI, MapF, TabI, TabF, T2).

check(M0, K, V, M) :- rb_lookup(K, V, M0), !, M = M0.
check(M0, K, V, M) :- rb_insert(M0, K, V, M).

increment_ref_count(V)  :-
	arg(1,V,I0),
	I is I0+1,
	setarg(1,V,I).
